/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/reiserfs_fs.h>
#include <linux/locks.h>
#include <asm/uaccess.h>

#else

#include "nokernel.h"

#endif


void inode2sd (struct stat_data * sd, struct inode * inode)
{
  sd->sd_mode = cpu_to_le16 (inode->i_mode);
  sd->sd_uid = cpu_to_le16 (inode->i_uid);
  sd->sd_gid = cpu_to_le16 (inode->i_gid);
  sd->sd_nlink = cpu_to_le16 (inode->i_nlink);
  sd->sd_size = cpu_to_le32 (inode->i_size);
  sd->sd_atime = cpu_to_le32 (inode->i_atime);
  sd->sd_ctime = cpu_to_le32 (inode->i_ctime);
  sd->sd_mtime = cpu_to_le32 (inode->i_mtime);
  if (S_ISREG (inode->i_mode))
    sd->u.first_direct_byte = cpu_to_le32 (inode->u.reiserfs_i.i_first_direct_byte);
  else if (S_ISDIR (inode->i_mode))
    sd->u.parent_dir_id = cpu_to_le32 (inode->u.reiserfs_i.i_parent_dir_id);
  else if (S_ISCHR (inode->i_mode) || S_ISBLK (inode->i_mode))
    sd->u.rdev = cpu_to_le32(kdev_t_to_nr (inode->i_rdev));
}


static void sd2inode (struct inode * inode, struct stat_data * sd)
{
  inode->i_mode = le16_to_cpu (sd->sd_mode);
  inode->i_uid = le16_to_cpu (sd->sd_uid);
  inode->i_gid = le16_to_cpu (sd->sd_gid);
  inode->i_nlink = le16_to_cpu (sd->sd_nlink);
  inode->i_size = le32_to_cpu (sd->sd_size);
  inode->i_mtime = le32_to_cpu (sd->sd_mtime);
  inode->i_atime = le32_to_cpu (sd->sd_atime);
  inode->i_ctime = le32_to_cpu (sd->sd_ctime);
  if (S_ISREG (inode->i_mode))
    inode->u.reiserfs_i.i_first_direct_byte = le32_to_cpu (sd->u.first_direct_byte);
  else if (S_ISDIR (inode->i_mode))
    inode->u.reiserfs_i.i_parent_dir_id = le32_to_cpu (sd->u.parent_dir_id);
  else if (S_ISCHR (inode->i_mode) || S_ISBLK (inode->i_mode))
    inode->i_rdev = to_kdev_t (le32_to_cpu (sd->u.rdev));
}



void reiserfs_delete_inode (struct inode * inode)
{
  /* The = 0 happens when we abort creating a new inode for some reason like lack of space.. */
  if (INODE_PKEY(inode)->file_id != 0) {
    reiserfs_delete_object (inode);
    reiserfs_release_objectid (inode->i_ino, inode->i_sb);
  } else
    /* no object items are in the tree */
    ;
  clear_inode (inode);
}


/* convert logical file block to appropriate unformatted node. */
int reiserfs_bmap (struct inode * inode, int block)
{
  struct key offset_key;
  struct path path_to_blocknr;
  int pos_in_item;
  int repeat;
  struct buffer_head * bh;
  struct item_head * ih;
  int blocknr;

  copy_key (&offset_key, INODE_PKEY (inode));
  offset_key.k_offset = block * inode->i_sb->s_blocksize + 1;
  if (INODE_OFFSET_IN_DIRECT (inode, offset_key.k_offset)) {
    reiserfs_warning ("vs-13010: reiserfs_bmap: %d byte is in direct item", offset_key.k_offset);
    return 0;
  }

  /*
  if (block < REISERFS_N_BLOCKS && inode->u.reiserfs_i.i_data [block]) {
    inode->i_sb->u.reiserfs_sb.s_bmaps_without_search ++;
    return inode->u.reiserfs_i.i_data [block];
  }
  */

  mark_key_indirect (&offset_key);
  init_path (&path_to_blocknr);
  if (search_for_position_by_key (inode->i_sb, &offset_key, &path_to_blocknr, &pos_in_item, &repeat) == POSITION_NOT_FOUND) {
    reiserfs_warning ("vs-13020: reiserfs_bmap: there is no required byte (%k) in the file of size %ld. Found item \n%h\n", 
		      &offset_key, inode->i_size, PATH_PITEM_HEAD (&path_to_blocknr));
    pathrelse (&path_to_blocknr);
    return 0;
  }

  bh = PATH_PLAST_BUFFER (&path_to_blocknr);
  ih = B_N_PITEM_HEAD (bh, PATH_LAST_POSITION (&path_to_blocknr));

  if (I_IS_INDIRECT_ITEM (ih)) {
    __u32 * ind_item = (__u32 *)B_I_PITEM (bh, ih);

    /*
    copy_data_blocks_to_inode (inode, ih, ind_item);
    */
    blocknr = ind_item [pos_in_item];
    pathrelse (&path_to_blocknr);
    inode->i_sb->u.reiserfs_sb.s_bmaps ++;
    return blocknr;
  }

  reiserfs_warning ("vs-13030: reiserfs_bmap: found item \n%h\n is not indirect one\n", ih);
  pathrelse (&path_to_blocknr);
  return 0;
}


static int read_file_tail (struct inode * inode, loff_t offset, char * buf, int size)
{
  struct key key;
  int repeat, pos_in_item;
  struct path path;
  struct item_head * ih;
  char * local_buf, * p;
  int chars, left, read;
  struct buffer_head * bh;

  if (!INODE_OFFSET_IN_DIRECT (inode, offset + 1)) {
    reiserfs_warning ("vs-18010: read_direct_item: given offset (%d) is not stored in direct item. \
Inode's first direct byte %d\n", offset, inode->u.reiserfs_i.i_first_direct_byte);
    return 0;
  }

  local_buf = kmalloc (size, GFP_KERNEL);
  if (local_buf == 0) {
    return -ENOMEM;
  }
  p = local_buf;

  path.path_length = ILLEGAL_PATH_ELEMENT_OFFSET;

  copy_key (&key, INODE_PKEY(inode));
  key.k_offset = offset + 1;

  left = size;
  read = 0;
  while (left) {
    if (search_for_position_by_key (inode->i_sb, &key, &path, &pos_in_item, &repeat) == POSITION_NOT_FOUND)
      break;

    bh = PATH_PLAST_BUFFER (&path);
    ih = B_N_PITEM_HEAD (bh, PATH_LAST_POSITION (&path));

    chars = ih->ih_item_len - pos_in_item;
    if (chars > left)
      chars = left;

    memcpy (p, B_I_PITEM (bh, ih) + pos_in_item, chars);
    key.k_offset += chars;
    read += chars;
    left -= chars;
    p += chars;
  }

  copy_to_user (buf, local_buf, read);
  kfree (local_buf);
  pathrelse (&path);
  return read;
}


int reiserfs_readpage (struct file * file, struct page * page) 
{
  struct inode * inode;

  inode = file->f_dentry->d_inode;

  if (page->offset == inode->u.reiserfs_i.i_first_direct_byte - 1) {
    /*atomic_inc(&page->count);*/
    memset ((char *)page_address (page), 0, PAGE_SIZE);
    read_file_tail (inode, page->offset, (char *)page_address (page), PAGE_SIZE);
    set_bit (PG_uptodate, &page->flags);
    return 0;
  }  else
    return generic_readpage (file, page);
}


/* iget accepts only super block and inode number as it hashes inodes using device identifier and
   inode number. If iget could not find required inode in its hash queues, then it calls
   reiserfs_read_inode passing to it only inode number. Reiserfs_read_inode must know the key. That
   is why we keep key in global array before iget. Stat data can be stored together with directory
   entry. Path_to_entry points there. Reiserfs_read_inode finds stat data on the path first, and
   uses search_by_key if there were no stat data.  */

#define KEY_ARRAY_SIZE 100

static struct stored_key {
  struct key sd_key;		/* key to look for stat data of the object */
  struct path * path_to_entry;	/* path to stat data */
} g_key_array [KEY_ARRAY_SIZE] = {{{{0,},},},};


/* store key of stat data (it can be entry item key and fair stat data key) */
/* path is path to item, that is directory entry or reset path.
   key is key of object pointed by directory entry (its first two components) */
static void store_key (struct path * path, struct key * key)
{
  int i;

  for (i = 0; i < sizeof (g_key_array) / sizeof (g_key_array[0]); i ++) {
    if (g_key_array[i].sd_key.file_id == 0) {
      copy_key (&(g_key_array[i].sd_key), key);
      g_key_array[i].path_to_entry = path;
      return;
    }
  }
  reiserfs_warning ("vs-13010: store_key: table of keys is full\n");
}


static struct stored_key * look_for_key  (unsigned long objectid)
{
  int i;

  for (i = 0; i < sizeof (g_key_array) / sizeof (g_key_array[0]); i ++) {
    if (g_key_array[i].sd_key.file_id == objectid)
      return g_key_array + i;
  }
  return 0;
}


static void forget_key (struct key * key)
{
  int i;

  for (i = 0; i < sizeof (g_key_array) / sizeof (g_key_array[0]); i ++) {
    if (key->file_id == g_key_array[i].sd_key.file_id) {
      g_key_array[i].sd_key.file_id = 0;
      return;
    }
  }
  reiserfs_warning ("vs-13015: forget_key: could not find key [%k] in the table\n", key);
}


static int last_path_element_correct (struct path * path)
{
  /* last buffer can be not in tree already. But if it is in tree, it must be leaf node */
  if (path->path_length != ILLEGAL_PATH_ELEMENT_OFFSET &&	/* path is not reset */
      B_IS_IN_TREE (PATH_PLAST_BUFFER (path)) &&		/* last buffer is in the path */
      (unsigned)PATH_LAST_POSITION (path) < B_NR_ITEMS (PATH_PLAST_BUFFER (path)))/* last position points to item */
    return 1;
  return 0;
}


static struct stat_data * look_for_sd (struct stored_key * skey, struct inode * inode)
{
  struct buffer_head * bh;
  struct item_head * ih;
  int repeat;
  
  
  if (!last_path_element_correct (skey->path_to_entry) || /* path not initialised or path has been corrupted */
      comp_keys (&(skey->sd_key), &(PATH_PITEM_HEAD (skey->path_to_entry)->ih_key)) != 0 /* entry item has been moved */ 
      )
    if (search_by_key (inode->i_sb, &(skey->sd_key), skey->path_to_entry, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) != ITEM_FOUND)
      return 0;
    
  bh = PATH_PLAST_BUFFER (skey->path_to_entry);
  ih = PATH_PITEM_HEAD (skey->path_to_entry);

  /* key of stat data */
  copy_key (INODE_PKEY (inode), &(skey->sd_key));
  
  return B_I_STAT_DATA (bh, ih);
}


/* looks for stat data in the tree, and fills up the fields of in-core
   inode stat data fields */
void reiserfs_read_inode (struct inode * inode)
{
  struct stat_data * sd;
  struct stored_key * skey;

  inode->i_op = NULL;
  inode->i_mode = 0;

  /* look for key of the object's stat data. It must be saved by reiserfs_iget before calling iget */
  skey = look_for_key (inode->i_ino);
  if (skey == 0) {
    reiserfs_warning ("vs-13025: reiserfs_read_inode: key stored by reiserfs_iget for object %d in above not found", inode->i_ino);
    return;
  }

  /* look for stat data. Re-search if necessary. Store 2 first components of object's items and key of object's stat data */
  sd = look_for_sd (skey, inode);
  if (sd == 0) {
    reiserfs_warning ("vs-13030: reiserfs_read_inode: stat data of object %d not found\n", inode->i_ino);
    return;
  }
  sd2inode (inode, sd);
  inode->i_blocks = inode->i_blksize = 0;

  if (S_ISREG (inode->i_mode))
    inode->i_op = &reiserfs_file_inode_operations;
  else if (S_ISDIR (inode->i_mode))
    inode->i_op = &reiserfs_dir_inode_operations;
  else if (S_ISLNK (inode->i_mode))
    inode->i_op = &reiserfs_symlink_inode_operations;
  else if (S_ISCHR (inode->i_mode))
    inode->i_op = &chrdev_inode_operations;
  else if (S_ISBLK (inode->i_mode))
    inode->i_op = &blkdev_inode_operations;
  else if (S_ISFIFO (inode->i_mode))
    init_fifo (inode);
}


/* key is key of object, path is path to directory entry */
struct inode * reiserfs_iget (struct super_block * s, struct path * path, struct key * key)
{
  struct inode * inode;

  store_key (path, key);
  inode = iget (s, key->file_id);
  forget_key (key);
  pathrelse (path);
  return inode;
}


static struct buffer_head * reiserfs_update_inode (struct inode * inode, struct path * path_to_sd, int read_blocks)
{
  struct stat_data * sd;
  int repeat;
  struct buffer_head * bh;
/*    
  if ((inode->i_nlink == 0 && comp_short_keys (INODE_PKEY (inode), INODE_SD_KEY (inode))) ||
      INODE_SD_KEY (inode)->k_dir_id == 0)
    return 0;
*/

  /* look for the object's stat data */
  if (search_by_key (inode->i_sb, INODE_PKEY (inode), path_to_sd, &repeat, DISK_LEAF_NODE_LEVEL, read_blocks) == ITEM_NOT_FOUND) {
    if (read_blocks != DONT_READ_BLOCKS) {
      /* this is called not from if_in_ram_update_sd */
      reiserfs_warning("vs-13035: reiserfs_update_inode: stat data of object [%k] not found\n", INODE_PKEY (inode));
    }
    return 0;
  }

  bh = PATH_PLAST_BUFFER (path_to_sd);
  sd = B_N_STAT_DATA (bh, PATH_LAST_POSITION (path_to_sd));

  /* do stat data update */
  inode2sd (sd, inode);

  mark_buffer_dirty (bh, 1);
  return bh;
}


/* looks for stat data, then copies fields to it, marks the buffer
   containing stat data as dirty */
void reiserfs_write_inode (struct inode * inode)
{
  struct path path_to_sd;

  init_path (&path_to_sd);
  reiserfs_update_inode (inode, &path_to_sd, READ_BLOCKS);
  pathrelse (&path_to_sd);
}


int reiserfs_sync_inode (struct inode * inode)
{
  int err = 0;
  struct path path_to_sd;
  struct buffer_head * bh;

  init_path (&path_to_sd);
  bh = reiserfs_update_inode (inode, &path_to_sd, READ_BLOCKS);
  if (bh && buffer_dirty (bh)) {
    ll_rw_block(WRITE, 1, &bh);
    wait_on_buffer(bh);
    if (buffer_req(bh) && !buffer_uptodate(bh)) {
      reiserfs_warning ("vs-13040: reiserfs_sync_inode: IO error syncing reiserfs stat data ["
	      "device:\"%s\", blocknr:%lu\n",
	      kdevname(inode->i_dev), bh->b_blocknr);
      err = -1;
    }
  } else {
    if (bh == 0)
      err = -1;
  }

  pathrelse (&path_to_sd);
  return err;
	
}


inline static void init_inode (const struct inode * dir, struct inode * inode)
{
  inode->i_sb = dir->i_sb;
  inode->i_flags = 0;

  inode->i_ino = reiserfs_get_unused_objectid (dir->i_sb);
  inode->i_nlink = 1;
  inode->i_dev = dir->i_sb->s_dev;
/*  inode->i_uid = current->fsuid;*/
  inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : 1/*current->fsgid*/;
  inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
  inode->i_op = NULL;
  inode->i_blocks = inode->i_blksize = 0;

}


/* init item head to be an item head of given type */
void init_item_head (struct item_head * ih, struct inode * inode, __u32 offset, int item_len, int type_bit)
{
  ih->ih_key.group_locale = INODE_PKEY (inode)->group_locale;
  ih->ih_key.name_hash = INODE_PKEY (inode)->name_hash;
  ih->ih_key.file_id = INODE_PKEY (inode)->file_id;
  ih->ih_key.k_offset = offset;
  ih->ih_free_space = 0xffff;
  ih->ih_type = 0;
  set_bit (type_bit, &(ih->ih_type));

  ih->ih_item_len = item_len;
}


struct inode * reiserfs_new_inode (const struct inode * dir)
{
  struct inode * inode;

  if (!dir || !(inode = get_empty_inode()))
    return 0;

  init_inode (dir, inode);

  INODE_PKEY (inode)->group_locale.major = dir->i_ino;
  INODE_PKEY (inode)->group_locale.minor = 0;
  INODE_PKEY (inode)->name_hash = 0;/* will be set in add_entry */
  INODE_PKEY (inode)->file_id = inode->i_ino;
  INODE_PKEY (inode)->k_offset = 0;
  inode->u.reiserfs_i.i_first_direct_byte = NO_BYTES_IN_DIRECT_ITEM;
  insert_inode_hash (inode);
  return inode;
}


struct inode * reiserfs_new_directory_inode (const struct inode * dir, int mode)
{
  struct item_head ih;
  struct stat_data sd;
  struct path path;
  struct inode * inode;
  int repeat;

  init_path (&path);

  if (!dir || !(inode = get_empty_inode()))
    return 0;

  init_inode (dir, inode);

  mode = S_IFDIR | (mode & 0777/* & ~current->fs->umask*/);
  if (dir->i_mode & S_ISGID)
    mode |= S_ISGID;
  inode->i_mode = mode;
  inode->i_nlink = 2;
  inode->i_size = EMPTY_DIR_SIZE;

  /* prepare item header */
  ih.ih_key.group_locale.major = inode->i_ino;
  ih.ih_key.group_locale.minor = 0;
  ih.ih_key.name_hash = 0;
  ih.ih_key.file_id = inode->i_ino;
  ih.ih_key.k_offset = 0;
  mark_stat_data (&ih);
  ih.ih_item_len = SD_SIZE;
  ih.ih_free_space = 0;
  
  /* initialize stat data */
  inode2sd (&sd, inode);
  sd.u.parent_dir_id = dir->i_ino;

  /* look for insert point for the directory's stat data */
  if (search_by_key (dir->i_sb, &(ih.ih_key), &path, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_FOUND) {
    reiserfs_warning ("vs-13045: reiserfs_new_directory_inode: item with this key (%k) already exists\n", &(ih.ih_key));
    iput (inode);
    return 0;
  }

  if (reiserfs_insert_item (dir->i_sb, &path, &ih, (char *)(&sd), 0/* zero number*/, NOTHING_SPECIAL) != OK) {
    iput (inode);
    return 0;
  }

  /* directory stat data has been inserted. */
  INODE_PKEY (inode)->group_locale.major = inode->i_ino;
  INODE_PKEY (inode)->group_locale.minor = 0;
  INODE_PKEY (inode)->name_hash = 0;
  INODE_PKEY (inode)->file_id = inode->i_ino;
  INODE_PKEY (inode)->k_offset = 0;
  inode->u.reiserfs_i.i_parent_dir_id = dir->i_ino;

  insert_inode_hash (inode);
  return inode;
}

#if 0
/* insert symname as direct item */
struct inode * reiserfs_new_symlink_inode (const struct inode * dir, const char * symname)
{
  struct inode * inode;
  struct super_block * sb;
  struct item_head ih;
  int repeat;
  struct path path;

  init_path (&path);

  if (!dir || !(inode = get_empty_inode()))
    return NULL;
  sb = dir->i_sb;

  init_inode (dir, inode);
  init_item_head (&ih, inode, 1/* offset */, strlen (symname), K_Direct_item);

  if (search_by_key (sb, &(ih.ih_key), &path, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_FOUND) {
    reiserfs_warning ("vs-13050: reiserfs_new_symlink_inode: item with this key (%k) already exists\n", &(ih.ih_key));
    iput (inode);
    return 0;
  }
  if (reiserfs_insert_item (sb, &path, &ih, symname, 0/* zero number*/, NOTHING_SPECIAL) != OK) {
    iput (inode);
    return 0;
  }

  /* symlink body has been inserted. */
  INODE_PKEY (inode)->pack_locale = inode->i_ino;
  INODE_PKEY (inode)->name_hash = 0;
  INODE_PKEY (inode)->file_id = inode->i_ino;
  INODE_PKEY (inode)->k_offset = 0;
  inode->u.reiserfs_i.i_first_direct_byte = 1;

  insert_inode_hash (inode);
  return inode;
}
#endif


#ifndef DIRTY_LATER
extern spinlock_t inode_lock;
#endif
                                /* This is the problem we are trying to solve: with reiserfs it is a bad thing to defer
                                an update of stat data if the buffer holding the stat data is in RAM. If we defer it,
                                then when we later do the update, the buffer might not still be in RAM.  Consider that
                                inodes have a longer lifetime in RAM, and you can see that this is especially important.

                                Unfortunately, sometimes we need to pass an inode, sometimes an iattr, sometimes just
                                one field, to the function that updates the stat data.  Umpteen different functions that
                                do essentially the same thing.  For now I implement just a straight replacement for
                                mark_inode_dirty -Hans */
                                /* This function is inefficient if only one field of the inode was changed. 
                                 This function can cause schedule.  */
void if_in_ram_update_sd (struct inode * inode)
{

#ifdef DIRTY_LATER

  mark_inode_dirty(inode);
  return;

#else /* DIRTY_LATER */

  extern struct list_head inode_in_use;
  struct path path_to_sd;

  init_path (&path_to_sd);

  
  if (!inode->i_nlink || (S_ISFIFO (inode->i_mode))) { 
    mark_inode_dirty(inode);    /* let us deferr the update since in this case we always follow this with an iput which
                                   will do the update (that is to say, the removal) of the stat data */
    return;
  }

#ifdef REISERFS_CHECK
  if (inode->i_sb) {
#endif /* REISERFS_CHECK */

    /* wait until inode is not locked and then lock it */
    while ((inode->i_state & I_LOCK)) {
      __wait_on_inode(inode);
    }
    inode->i_state |=  I_LOCK;

    if (reiserfs_update_inode (inode, &path_to_sd, DONT_READ_BLOCKS)) {
      spin_lock(&inode_lock);   /* locks inode lists (not the inode) */
      if ((inode->i_state & I_DIRTY)) {
        struct list_head *insert = &inode_in_use;

        /* mark inode clean and take it off dirty list */
        inode->i_state &= ~I_DIRTY;
#ifdef REISERFS_CHECK
        /* Only add valid (ie hashed) inodes to the in_use list */
        if (!list_empty(&inode->i_hash)) {
#endif /* REISERFS_CHECK */
          list_del(&inode->i_list);
          list_add(&inode->i_list, insert);
#ifdef REISERFS_CHECK
        } 
        else 
          printk("reiser-1805: if_in_ram_update_sd: a dirty inode was not on any inode list\n");
#endif /* REISERFS_CHECK */
      }
      spin_unlock(&inode_lock);
      /*mark_buffer_dirty (bh, 1);*/
    } else {   /* stat data item for this inode no longer in RAM */
      /* if stat data was not found then it must be because it is no longer in RAM, so just mark it
	 dirty for now, and let somebody else write it */
      mark_inode_dirty(inode);
    }  
    inode->i_state &= ~I_LOCK;
    wake_up(&inode->i_wait);
    pathrelse(&path_to_sd);
#ifdef REISERFS_CHECK
  }
  else
    printk("reiser-1804: if_in_ram_update_sd: !sb, should not happen\n");
#endif /* REISERFS_CHECK */

#endif /* ! DIRTY_LATER */

}

#ifdef __KERNEL__

/* this generates little if any speedup compared to if_in_ram_update_sd(), might not be worthwhile
   code, probably search_by_key dominates cpu consumption */
void if_in_ram_update_some_sd (struct inode * inode,  struct iattr * attr)
{
  struct stat_data * sd;
  int repeat;
  struct buffer_head * bh;
  extern struct list_head inode_in_use;
  unsigned int ia_valid = attr->ia_valid;
  struct path path_to_sd;

  init_path (&path_to_sd);

  if (!inode->i_nlink || (S_ISFIFO (inode->i_mode))) { /* Vladimir, what do you think, is the S_FIFO needed here?-Hans */
    mark_inode_dirty(inode);    /* let us deferr the update since in this case we always follow this with an iput which
                                   will do the update (that is to say, the removal) of the stat data */
    return;
  }

#ifdef REISERFS_CHECK
  if (inode->i_sb) {
#endif /* REISERFS_CHECK */

                                /* wait until inode is not locked and then lock it */
    while ((inode->i_state & I_LOCK)) {
      __wait_on_inode(inode);
    }
    inode->i_state |=  I_LOCK;  /* depends on interrupts never locking inodes, I suppose it is correct.... */

    /* look for the stat data item, if we find it in RAM, sync to it now, else put the inode on the list of dirty inodes
       to be synced someday using mark_inode_dirty and perhaps the sync will happen when the buffer is back in RAM or
       after a bunch of other changes to it have been successfully cached by deferring the I/O */
    if (search_by_key (inode->i_sb, INODE_PKEY (inode), &path_to_sd, &repeat, DISK_LEAF_NODE_LEVEL, DONT_READ_BLOCKS) == ITEM_FOUND) {
                                /* all of these branches are very inefficient, but the search_by_key is probably the
                                   worst of it */
      bh = PATH_PLAST_BUFFER (&path_to_sd);
      sd = B_N_STAT_DATA (bh, PATH_LAST_POSITION (&path_to_sd));
        if (ia_valid & ATTR_MODE)
          sd->sd_mode = cpu_to_le16 (inode->i_mode);
        if (ia_valid & ATTR_UID)
          sd->sd_uid = cpu_to_le16 (inode->i_uid);
        if (ia_valid & ATTR_GID)
          sd->sd_gid = cpu_to_le16 (inode->i_gid);
        if (ia_valid & ATTR_SIZE)
          sd->sd_size = cpu_to_le32 (inode->i_size);
        if (ia_valid & ATTR_MTIME)
          sd->sd_mtime = cpu_to_le32 (inode->i_mtime);
        if (ia_valid & ATTR_CTIME)
          sd->sd_ctime = cpu_to_le32 (inode->i_ctime);
      spin_lock(&inode_lock);   /* locks inode lists (not the inode) */
      if ((inode->i_state & I_DIRTY)) {
        struct list_head *insert = &inode_in_use;

        /*      mark inode clean and take it off dirty list */
        inode->i_state &= ~I_DIRTY;
#ifdef REISERFS_CHECK
        /* Only add valid (ie hashed) inodes to the in_use list */
        if (!list_empty(&inode->i_hash)) {
#endif /* REISERFS_CHECK */
          list_del(&inode->i_list);
          list_add(&inode->i_list, insert);
#ifdef REISERFS_CHECK
        }
        else 
          printk("reiser-1806: if_in_ram_update_some_sd: a dirty inode was not on any inode list, maybe a pipe?");
#endif /* REISERFS_CHECK */
      }
      spin_unlock(&inode_lock);
      mark_buffer_dirty (bh, 1);
    } else {   /* stat data item for this inode no longer in RAM */
      inode->i_state &= ~I_LOCK; /* probably should combine two lines that unlock inode by postponing the unlock.... */
                                /* if inode was not found then it must be because it is no longer in RAM, so just
                                   mark it dirty for now, and let somebody else write it */
      mark_inode_dirty(inode);
    }  
    inode->i_state &= ~I_LOCK;
    wake_up(&inode->i_wait);
    pathrelse(&path_to_sd);
#ifdef REISERFS_CHECK
  }
  else
    printk("reiser-1807: if_in_ram_update_some_sd: !sb, should not happen");
#endif /* REISERFS_CHECK */
}


void inline reiserfs_inode_setattr(struct inode * inode, struct iattr * attr)
{
        unsigned int ia_valid = attr->ia_valid;

        if (ia_valid & ATTR_UID)
                inode->i_uid = attr->ia_uid;
        if (ia_valid & ATTR_GID)
                inode->i_gid = attr->ia_gid;
        if (ia_valid & ATTR_SIZE)
                inode->i_size = attr->ia_size;
        if (ia_valid & ATTR_ATIME)
                inode->i_atime = attr->ia_atime;
        if (ia_valid & ATTR_MTIME)
                inode->i_mtime = attr->ia_mtime;
        if (ia_valid & ATTR_CTIME)
                inode->i_ctime = attr->ia_ctime;
        if (ia_valid & ATTR_MODE) {
                inode->i_mode = attr->ia_mode;
                if (!in_group_p(inode->i_gid) && !capable(CAP_FSETID))
                        inode->i_mode &= ~S_ISGID;
        }
#ifdef DIRTY_LATER
/*      printk("active 2"); */
        mark_inode_dirty(inode);
#else
        if_in_ram_update_some_sd (inode, attr);
#endif
}


int reiserfs_notify_change(struct dentry * dentry, struct iattr * attr)
{
        struct inode *inode = dentry->d_inode;
        int error;

        error = inode_change_ok(inode, attr);
        if (!error)
          reiserfs_inode_setattr(inode, attr);
        return error;
}

/* I believe that further optimizing this code is best done by optimizing search_by_key.. */


#endif /* __KERNEL__ */














